@@ -53,12 +53,12 @@
 #ifndef QDBUSCONNECTION_P_H
 #define QDBUSCONNECTION_P_H
 
+#include <QtDBus/private/qtdbusglobal_p.h>
 #include <qdbuserror.h>
 #include <qdbusconnection.h>
 
 #include <QtCore/qatomic.h>
 #include <QtCore/qhash.h>
-#include <QtCore/qmutex.h>
 #include <QtCore/qobject.h>
 #include <QtCore/qpointer.h>
 #include <QtCore/qreadwritelock.h>
@@ -94,7 +94,7 @@ class QDBusServer;
 class QDBusErrorInternal
 {
     mutable DBusError error;
-    Q_DISABLE_COPY(QDBusErrorInternal)
+    Q_DISABLE_COPY_MOVE(QDBusErrorInternal)
 public:
     inline QDBusErrorInternal() { q_dbus_error_init(&error); }
     inline ~QDBusErrorInternal() { q_dbus_error_free(&error); }
@@ -106,7 +106,7 @@ public:
 // QDBusConnectionPrivate holds the DBusConnection and
 // can have many QDBusConnection objects referring to it
 
-class QDBusConnectionPrivate: public QObject
+class Q_AUTOTEST_EXPORT QDBusConnectionPrivate: public QObject
 {
     Q_OBJECT
 public:
@@ -121,6 +121,15 @@ public:
         QSocketNotifier *write;
     };
 
+    struct ArgMatchRules {
+        QStringList args;
+        QString arg0namespace;
+        bool operator==(const ArgMatchRules &other) const {
+            return args == other.args &&
+                   arg0namespace == other.arg0namespace;
+        }
+    };
+
     struct SignalHook
     {
         inline SignalHook() : obj(0), midx(-1) { }
@@ -128,7 +137,7 @@ public:
         QObject* obj;
         int midx;
         QVector<int> params;
-        QStringList argumentMatch;
+        ArgMatchRules argumentMatch;
         QByteArray matchRule;
     };
 
@@ -148,14 +157,6 @@ public:
             { return name < other; }
         inline bool operator<(const QStringRef &other) const
             { return QStringRef(&name) < other; }
-#if defined(Q_CC_MSVC) && _MSC_VER < 1600
-        inline bool operator<(const ObjectTreeNode &other) const
-            { return name < other.name; }
-        friend inline bool operator<(const QString &str, const ObjectTreeNode &obj)
-            { return str < obj.name; }
-        friend inline bool operator<(const QStringRef &str, const ObjectTreeNode &obj)
-            { return str < QStringRef(&obj.name); }
-#endif
         inline bool isActive() const
         { return obj || !children.isEmpty(); }
 
@@ -197,7 +198,6 @@ public:
     ~QDBusConnectionPrivate();
 
     void createBusService();
-    void setDispatchEnabled(bool enable);
     void setPeer(DBusConnection *connection, const QDBusErrorInternal &error);
     void setConnection(DBusConnection *connection, const QDBusErrorInternal &error);
     void setServer(QDBusServer *object, DBusServer *server, const QDBusErrorInternal &error);
@@ -216,12 +216,19 @@ public:
     QDBusMessage sendWithReplyLocal(const QDBusMessage &message);
     QDBusPendingCallPrivate *sendWithReplyAsync(const QDBusMessage &message, QObject *receiver,
                                                 const char *returnMethod, const char *errorMethod,int timeout = -1);
+
     bool connectSignal(const QString &service, const QString &path, const QString& interface,
                        const QString &name, const QStringList &argumentMatch, const QString &signature,
                        QObject *receiver, const char *slot);
     bool disconnectSignal(const QString &service, const QString &path, const QString& interface,
                           const QString &name, const QStringList &argumentMatch, const QString &signature,
                           QObject *receiver, const char *slot);
+    bool connectSignal(const QString &service, const QString &path, const QString& interface,
+                       const QString &name, const ArgMatchRules &argumentMatch, const QString &signature,
+                       QObject *receiver, const char *slot);
+    bool disconnectSignal(const QString &service, const QString &path, const QString& interface,
+                          const QString &name, const ArgMatchRules &argumentMatch, const QString &signature,
+                          QObject *receiver, const char *slot);
     void registerObject(const ObjectTreeNode *node);
     void unregisterObject(const QString &path, QDBusConnection::UnregisterMode mode);
     void connectRelay(const QString &service,
@@ -270,16 +277,17 @@ private:
     void _q_newConnection(QDBusConnectionPrivate *newConnection);
 
 protected:
-    void timerEvent(QTimerEvent *e) Q_DECL_OVERRIDE;
+    void timerEvent(QTimerEvent *e) override;
 
 public slots:
     // public slots
+    void setDispatchEnabled(bool enable);
     void doDispatch();
     void socketRead(int);
     void socketWrite(int);
     void objectDestroyed(QObject *o);
     void relaySignal(QObject *obj, const QMetaObject *, int signalId, const QVariantList &args);
-    void addSignalHook(const QString &key, const SignalHook &hook);
+    bool addSignalHook(const QString &key, const SignalHook &hook);
     bool removeSignalHook(const QString &key, const SignalHook &hook);
 
 private slots:
@@ -292,7 +300,7 @@ signals:
     void dispatchStatusChanged();
     void spyHooksFinished(const QDBusMessage &msg);
     void messageNeedsSending(QDBusPendingCallPrivate *pcall, void *msg, int timeout = -1);
-    void signalNeedsConnecting(const QString &key, const QDBusConnectionPrivate::SignalHook &hook);
+    bool signalNeedsConnecting(const QString &key, const QDBusConnectionPrivate::SignalHook &hook);
     bool signalNeedsDisconnecting(const QString &key, const QDBusConnectionPrivate::SignalHook &hook);
     void serviceOwnerChanged(const QString &name, const QString &oldOwner, const QString &newOwner);
     void callWithCallbackFailed(const QDBusError &error, const QDBusMessage &message);
@@ -311,9 +319,6 @@ public:
         QDBusServer *serverObject;
     };
 
-    // the dispatch lock protects everything related to the DBusConnection or DBusServer
-    // including the timeouts and watches
-    QMutex dispatchLock;
     union {
         DBusConnection *connection;
         DBusServer *server;
@@ -343,13 +348,10 @@ public:
     static bool prepareHook(QDBusConnectionPrivate::SignalHook &hook, QString &key,
                             const QString &service,
                             const QString &path, const QString &interface, const QString &name,
-                            const QStringList &argMatch,
+                            const ArgMatchRules &argMatch,
                             QObject *receiver, const char *signal, int minMIdx,
                             bool buildSignature);
     static DBusHandlerResult messageFilter(DBusConnection *, DBusMessage *, void *);
-    static bool checkReplyForDelivery(QDBusConnectionPrivate *target, QObject *object,
-                                      int idx, const QList<int> &metaTypes,
-                                      const QDBusMessage &msg);
     static QDBusCallDeliveryEvent *prepareReply(QDBusConnectionPrivate *target, QObject *object,
                                                 int idx, const QVector<int> &metaTypes,
                                                 const QDBusMessage &msg);
@@ -380,6 +382,27 @@ extern QDBusMessage qDBusPropertySet(const QDBusConnectionPrivate::ObjectTreeNod
                                      const QDBusMessage &msg);
 extern QDBusMessage qDBusPropertyGetAll(const QDBusConnectionPrivate::ObjectTreeNode &node,
                                         const QDBusMessage &msg);
+
+// can be replaced with a lambda in Qt 5.7
+class QDBusConnectionDispatchEnabler : public QObject
+{
+    Q_OBJECT
+    QDBusConnectionPrivate *con;
+public:
+    QDBusConnectionDispatchEnabler(QDBusConnectionPrivate *con) : con(con) {}
+
+public slots:
+    void execute()
+    {
+        // This call cannot race with something disabling dispatch only because dispatch is
+        // never re-disabled from Qt code on an in-use connection once it has been enabled.
+        QMetaObject::invokeMethod(con, "setDispatchEnabled", Qt::QueuedConnection, Q_ARG(bool, true));
+        if (!con->ref.deref())
+            con->deleteLater();
+        deleteLater();
+    }
+};
+
 #endif // QT_BOOTSTRAPPED
 
 QT_END_NAMESPACE
